/***************/
/* L3Grilles.c */
/***************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "L3Environnement.h"
#include "L3Grilles.h"

#ifndef lint
    static const char sccs_id[] = "%W% %G% G. Colin de Verdiere";
#endif

/*      Modifs Pierre Brochard */

/*-
	Prototypes des fonctions internes
*/

static void
L3AxeGradue(
	    int sousimage,
	    REEL xmin,
	    REEL ymin,
	    REEL xmax,
	    REEL ymax,
	    int direction,
	    int nintervalles,
	    int xaxelog,
	    int yaxelog
);

static void
L3Graduation(
	     int sousimage,
	     REEL xutil,
	     REEL yutil,
	     int direction,
	     REEL longueur,
	     double taille,
	     int label,
	     int xaxelog,
	     int yaxelog,
	     int nsignif,
	     int decale,
	     double facteur,
 /* Encombrement du dernier label */
	     int *lastdx,
	     int *lastdy,
 /* Nombre de decimale des graduations */
	     int Impose,
	     int Principal
);

/*****************************************************************************
 * Procedure utilisateur : L3Axe                                             *
 *---------------------------------------------------------------------------*
 * Description : ajout d'un marqueur CODEL3AXE                               *
 *                                                                           *
 * Erreurs :    aucune                                                       *
 *                                                                           *
 *****************************************************************************/

void 
L3Axe(void)
{
	int codel3[2];

	codel3[0] = 2;
	codel3[1] = CODEL3AXE;

	L3AppelDrivers(codel3);
}

/***************************************
 * Procedure utilisateur : L3DefTitreAxe
 *--------------------------------------
 * Parametres
 *   sousimage : numero de la sous-image concernee
 *   nom_axe   : nom de l'axe (BAS, HAUT, GAUCHE, DROITE)
 *   titre     : pointeur sur la chaine de caracteres contenant le titre
 *   unite     : pointeur sur la chaine de caracteres contenant l'unite

 * Valeur de retour : aucune

 * Description : voir Doc

 * Erreurs
 *   axe incorrect
 *   titre incorrect
 *   unite incorrecte
 ******************************************************/

void
L3DefTitreAxe(
	      int sousimage,
	      int nom_axe,
	      char titre[],
	      char unite[]
)
{
   SETSOUSIMAGE(sousimage);
   if ((nom_axe != BAS) && (nom_axe != HAUT) && (nom_axe != GAUCHE) && (nom_axe != DROITE)) {
      fprintf(stderr, "L3DefTitreAxe : axe incorrect.\n");
   } else {
      if ((titre == NULL) || ((int) strlen(titre) > 255)) {
	 fprintf(stderr, "L3DefTitreAxe : titre incorrect.\n");
      } else {
	 strcpy(P_ParamSousImage->axe[nom_axe].titre, titre);
      }
      if ((unite == NULL) || ((int) strlen(unite) > 31)) {
	 fprintf(stderr, "L3DefTitreAxe : unite incorrecte.\n");
      } else {
	 strcpy(P_ParamSousImage->axe[nom_axe].unite, unite);
      }
   }
}

/**************************************
 * Procedure utilisateur : L3DefModeLog
 *-------------------------------------
 * Parametres
 *   sousimage : numero de la sous-image concernee
 *   xcoordlog : indicateur de mode log sur x
 *   ycoordlog : indicateur de mode log sur y

 * Valeur de retour : aucune

 * Description : voir Doc

 * Erreurs
 *   les indicateurs de mode logarithmique doivent valoir VRAI ou FAUX
 *   l'espace de coordonnees en x interdit le mode logarithmique
 *   l'espace de coordonnees en y interdit le mode logarithmique
 ******************************************************************/

void
L3DefModeLog(
	     int sousimage,
	     int xcoordlog,
	     int ycoordlog
)
{
   SETSOUSIMAGE(sousimage);
   if (((xcoordlog == VRAI) || (xcoordlog == FAUX)) && ((ycoordlog == VRAI) || (ycoordlog == FAUX))) {
      if ((XU <= 0.0) && (xcoordlog == VRAI)) {
	 fprintf(stderr, "L3DefModeLog : l'espace de coordonnees en x interdit le mode logarithmique.\n");
      } else if ((YU <= 0.0) && (ycoordlog == VRAI)) {
	 fprintf(stderr, "L3DefModeLog : l'espace de coordonnees en y interdit le mode logarithmique.\n");
      } else {

	 XCOORDLOG = xcoordlog;
	 YCOORDLOG = ycoordlog;
      }
   } else {
      fprintf(stderr, "L3DefModeLog : les indicateurs de mode logarithmique doivent valoir VRAI ou FAUX.\n");
   }
}

/*************************************
 * Procedure utilisateur : L3DefGrille
 *------------------------------------
 * Parametres
 *   sousimage : numero de la sous-image concernee
 *   axe       : nom de l'axe concerne (BAS, HAUT, GAUCHE ou DROITE)
 *   typeaxe   : type de l'axe concerne (AXE_PLEIN, AXE_CREUX, AXE_SEMICREUX,
 *                 AXE_VIDE, AXE_ABSENT, INCHANGE)
 *   label     : indicateur de presence des valeurs sur l'axe (VRAI, FAUX)
 *   gradinter : nombre de graduations intermediaires sur l'axe (lineaire)
 *               (AUTOMATIQUE, INCHANGE, 0 ou un nombre entier positif)
 *   factorisation : type de mise en facteur pour les labels (AUCUNE,
 *                   MINIMALE, MOYENNE, MAXIMALE, INCHANGE)

 * Valeur de retour : aucune

 * Description : voir Doc

 * Erreurs
 *    axe incorrect
 *    type d'axe incorrect
 *    indicateur de presence des valeurs incorrect
 *    nombre de graduations intermediaires incorrect
 *    type de factorisation incorrecte
 ****************************************************/

void
L3DefGrille(
	    int sousimage,
	    int axe,
	    int typeaxe,
	    int label,
	    int gradinter,
	    int factorisation
)
{
   SETSOUSIMAGE(sousimage);
   if ((axe != BAS) && (axe != HAUT) && (axe != GAUCHE) && (axe != DROITE)) {
      fprintf(stderr, "L3DefGrille : axe incorrect :%d\n", axe);
   } else {
      switch (typeaxe) {
       case AXE_PLEIN:
       case AXE_CREUX:
       case AXE_SEMICREUX:
       case AXE_VIDE:
       case AXE_ABSENT:
	 P_ParamSousImage->axe[axe].typeaxe = typeaxe;
	 break;
       case INCHANGE:
	 break;
       default:
	 fprintf(stderr,
			"L3DefGrille : type d'axe incorrect :%d\n", typeaxe);
      }

      switch (label) {
       case VRAI:
       case FAUX:
	 P_ParamSousImage->axe[axe].label = label;
	 break;
       case INCHANGE:
	 break;
       default:
	 fprintf(stderr,
	 "L3DefGrille : indicateur de presence des valeurs incorrect :%d\n",
			label);
      }

      if (gradinter != INCHANGE) {
	 if ((gradinter >= 0) || (gradinter == AUTOMATIQUE)) {
	    switch (axe) {
	     case BAS:
	     case HAUT:
	       P_ParamSousImage->axe[BAS].gradinter = gradinter;
	       P_ParamSousImage->axe[HAUT].gradinter = gradinter;
	       break;
	     case GAUCHE:
	     case DROITE:
	       P_ParamSousImage->axe[GAUCHE].gradinter = gradinter;
	       P_ParamSousImage->axe[DROITE].gradinter = gradinter;
	       break;
	    }
	 } else {
	    fprintf(stderr,
			   "L3DefGrille : nombre de graduations intermediaires incorrect :%d\n",
			   gradinter);
	 }
      }
      switch (factorisation) {
       case INCHANGE:
	 break;
       case AUCUNE:
       case MINIMALE:
       case MOYENNE:
       case MAXIMALE:
	 P_ParamSousImage->axe[axe].factorisation = factorisation;
	 break;
       default:
	 /*--------------------------------------------*
          * Erreur : type de factorisation incorrecte. *
          *--------------------------------------------*/

	 fprintf(stderr, "L3DefGrille : type de factorisation incorrecte :%d\n", factorisation);
      }
   }
}

/******************************************
 * Procedure utilisateur : L3RestaureGrille
 *-----------------------------------------
 * Parametres
 *   sousimage : numero de la sous-image concernee

 * Valeur de retour : aucune

 * Description : voir Doc

 * Erreurs : aucune
 *************************************************/

void
L3RestaureGrille(
		 int sousimage
)
{
   SETSOUSIMAGE(sousimage);

   TYPEAXEBAS = AXE_SEMICREUX;
   LABELBAS = VRAI;
   GRADINTERBAS = AUTOMATIQUE;
   FACTORISATIONBAS = AUCUNE;
   strcpy(TITREBAS, "");
   strcpy(UNITEBAS, "");

   TYPEAXEHAUT = AXE_CREUX;
   LABELHAUT = FAUX;
   GRADINTERHAUT = AUTOMATIQUE;
   FACTORISATIONHAUT = AUCUNE;
   strcpy(TITREHAUT, "");
   strcpy(UNITEHAUT, "");

   TYPEAXEGAUCHE = AXE_SEMICREUX;
   LABELGAUCHE = VRAI;
   GRADINTERGAUCHE = AUTOMATIQUE;
   FACTORISATIONGAUCHE = AUCUNE;
   strcpy(TITREGAUCHE, "");
   strcpy(UNITEGAUCHE, "");

   TYPEAXEDROITE = AXE_CREUX;
   LABELDROITE = FAUX;
   GRADINTERDROITE = AUTOMATIQUE;
   FACTORISATIONDROITE = AUCUNE;
   strcpy(TITREDROITE, "");
   strcpy(UNITEDROITE, "");
}

/***********************************
 * Procedure utilisateur : L3Grille
 *----------------------------------
 * Parametres
 *   sousimage : numero de la sous-image concernee
 *   x1, y1    : coordonnees utilisateur du coin inferieur gauche de la grille
 *   x2, y2    : coordonnees utilisateur du coin superieur droit de la grille
 *   nx        : nombre de graduations desire sur les axes horizontaux
 *   ny        : nombre de graduations desire sur les axes verticaux
 *   modes     : indicateur de modes logarithmiques (LINEAIRE, XAXLOG, YAXLOG,
 *                 XCOLOG, YCOLOG ou la somme d'un symbole en X et d'un en Y
 *                 parmi les quatres symboles precedents)

 * Valeur de retour : aucune

 * Description : voir Doc

 * Erreurs
 *    nombre de graduations incorrect
 **************************************************/

void
L3Grille(
		int sousimage,
		REEL x1,
		REEL y1,
		REEL x2,
		REEL y2,
		int nx,
		int ny,
		int modes
		)
{
	int             troncature;
	int             xcoordlog, ycoordlog;
	int             xaxelog, yaxelog;
	REEL            bx1, by1, bx2, by2;

	xcoordlog = (modes & XCOLOG) == XCOLOG;
	xaxelog = ((modes & XAXLOG) == XAXLOG) || xcoordlog;

	ycoordlog = (modes & YCOLOG) == YCOLOG;
	yaxelog = ((modes & YAXLOG) == YAXLOG) || ycoordlog;

	troncature = CLIPPING;

	if (troncature == VRAI)
	    L3Troncature(sousimage, CLIPPINGINACTIF);

    /* protection des grilles contre des bornes idiotes */
	if (xcoordlog) {
	    L3VerifieBorne(x1, x2, &bx1, &bx2);
	} else {
	    bx1 = x1;
	    bx2 = x2;
	    if (L3Egal(bx1, bx2, 1.e-5)) {
		bx1 *= .99;
		bx2 *= 1.01;
	    }
	    if (bx1 == bx2) {
		bx1 = -1.;
		bx2 = +1.;		/* cas de zero */
	    }
	    if (bx1 > bx2) {
		REEL b=bx2;
		bx2 = bx1;
		bx1 = b;
		fprintf(stderr,"L3Grilles bornes incoherentes "
				"retournees x1=%g x2=%g\n",bx1, bx2);
	    }
	}

	if (ycoordlog) {
	    L3VerifieBorne(y1, y2, &by1, &by2);
	} else {
	    by1 = y1;
	    by2 = y2;

	    if (L3Egal(by1, by2, 1.e-5)) {
		by1 *= .99;
		by2 *= 1.01;
	    }

	    if (by1 == by2) {
		by1 = -1.;
		by2 = +1.;		/* cas de zero */
	    }

	    if (by1 > by2) {
		REEL b=by2;
		by2 = by1;
		by1 = b;
		fprintf(stderr,"L3Grilles bornes incoherentes "
				"retournees y1=%g y2=%g\n",by1, by2);
	    }

	}

	L3DefSpaceLog(sousimage, bx1, by1, bx2, by2, xcoordlog, ycoordlog);
	L3AxeGradue(sousimage, bx1, by1, bx2, by2, BAS, nx, xaxelog, yaxelog);
	L3AxeGradue(sousimage, bx1, by1, bx2, by2, HAUT, nx, xaxelog, yaxelog);
	L3AxeGradue(sousimage, bx1, by1, bx2, by2, GAUCHE, ny, xaxelog, yaxelog);
	L3AxeGradue(sousimage, bx1, by1, bx2, by2, DROITE, ny, xaxelog, yaxelog);

	if (troncature == VRAI)
	    L3Troncature(sousimage, CLIPPINGACTIF);

}

/*********************************
 * Procedure interne : L3AxeGradue
 *--------------------------------
 * Parametres :
 *   sousimage    : numero de la sous-image concernee
 *   xmin, ymin   : coordonnees utilisateur du coin inferieur
 *                  gauche de la grille
 *   xmax, ymax   : coordonnees utilisateur du coin superieur
 *                  droit de la grille
 *   direction    : indicateur de l'axe a tracer
 *   nintervalles : nombre de graduations desire sur l'axe
 *   xaxelog      : indicateur de mode logarithmique sur x
 *   yaxelog      : indicateur de mode logarithmique sur y

 * Valeur de retour : aucune

 * Description : trace un axe gradue

 * Erreurs
 *    espace memoire insuffisant
 ***********************************************************/


static void
L3AxeGradue(
	    int sousimage,
	    REEL xmin,
	    REEL ymin,
	    REEL xmax,
	    REEL ymax,
	    int direction,
	    int nintervalles,
	    int xaxelog,
	    int yaxelog
)
{
   int             i;
   int             modelog = xaxelog;
   int             ninter;
   int             label;
   int             nsignif;
   int             decale;
   int             inter;
   int             gradinter;
   int             typeaxe;
   int             labelpresent;
   int             factorisation;
   int             xl3, yl3;
   int             position = POSMB;
   char           *titre;
   char           *unite;
   REEL            xdebut, ydebut, xfin, yfin;
   REEL            min = xmin, max = xmax, longueur = ymax - ymin;
   REEL            xutil, yutil;
   double          taille;
   double          angle = 0.;
   REEL            bidon;
   double          facteur = 1.0;
   GRADUATION     *tabvaleur;
   int             lgtabval;
   char            titrecomplet[320];
   char            chainefact[32];
   char            chainenormal[2];
   char            texte[256];
   /* Encombrement du dernier label trace pour eviter les recouvrements */
   int             lastdx = -LARGEURPAGEL3;
   int             lastdy = -LARGEURPAGEL3;
   int             maxdeci = -1;
   int             Principal = FAUX;

   SETSOUSIMAGE(sousimage);

   gradinter = P_ParamSousImage->axe[direction].gradinter;
   typeaxe = P_ParamSousImage->axe[direction].typeaxe;
   labelpresent = P_ParamSousImage->axe[direction].label;
   factorisation = P_ParamSousImage->axe[direction].factorisation;
   titre = P_ParamSousImage->axe[direction].titre;
   unite = P_ParamSousImage->axe[direction].unite;

   if (typeaxe != AXE_ABSENT)
   {
      /*----------------*
       * Trace de l'axe *
       *----------------*/

      switch (direction)
      {
       case BAS:
	 xdebut = xmin;
	 xfin = xmax;
	 ydebut = yfin = ymin;
	 min = xmin;
	 max = xmax;
	 longueur = ymax - ymin;
	 modelog = xaxelog;
	 break;
       case HAUT:
	 xdebut = xmin;
	 xfin = xmax;
	 ydebut = yfin = ymax;
	 min = xmin;
	 max = xmax;
	 longueur = ymax - ymin;
	 modelog = xaxelog;
	 break;
       case GAUCHE:
	 xdebut = xfin = xmin;
	 ydebut = ymin;
	 yfin = ymax;
	 min = ymin;
	 max = ymax;
	 longueur = xmax - xmin;
	 modelog = yaxelog;
	 break;
       case DROITE:
	 xdebut = xfin = xmax;
	 ydebut = ymin;
	 yfin = ymax;
	 min = ymin;
	 max = ymax;
	 longueur = xmax - xmin;
	 modelog = yaxelog;
	 break;
      }

      L3Axe();
      L3Segment(sousimage, &xdebut, &ydebut, &xfin, &yfin, 1);

      if (typeaxe != AXE_VIDE)
      {

	 if (gradinter > 0)
	    inter = gradinter + 2;
	 else
	    inter = 12;

	 lgtabval = (abs(nintervalles) + 3) * inter;

	 if ((tabvaleur = (GRADUATION*) malloc((unsigned int)
			 sizeof(GRADUATION) * lgtabval)) != NULL)
	 {
	    /*---------------------------*
	     * Determination des valeurs *
	     *---------------------------*/

	    L3CalculValeurs(min, max, nintervalles, modelog,
			    gradinter, labelpresent, typeaxe, factorisation,
			    &ninter, &lgtabval, &tabvaleur, &facteur);

	    /* Determination du nombre de decimales a imposer */
	    for (i = 0; i < ninter; i++)
	    {
	       int j, n, l, vu;

	       if (tabvaleur[i].label == FAUX)
		  continue;

	       if ((xaxelog && ((direction == BAS) || (direction == HAUT))) ||
			       (yaxelog && ((direction == DROITE) ||
					       (direction == GAUCHE))))
	       {
		  /* Cas d'un axe log -> notation scientifique imposee */
		  L3BeauLabelScientifique(tabvaleur[i].valeur,
					  tabvaleur[i].nsignif,
					  texte, -1);
	       }
	       else
	       {
		  /* cas lineaire pas de notation scientifique obligatoire */
		  L3BeauLabelPur(tabvaleur[i].valeur,
				 tabvaleur[i].nsignif,
				 texte, -1);
	       }

	       /* compte du nombre de decimales */
	       vu = FALSE;
	       l = strlen(texte);

	       for (j = 0, n = 0; j < l; j++)
	       {

		  if (texte[j] == '.')
		  {
		     vu = TRUE;
		     continue;
		  }

		  if (vu)
		  {

		     if (texte[j] == ' ') break;

		     n++;
		  }

	       }

	       maxdeci = (maxdeci < n) ? n : maxdeci;
	    }

	    /*-----------------------*
	     * Trace des graduations *
	     *-----------------------*/

	    for (i = 0; i < ninter; i++)
	    {

	       switch (direction)
	       {
		case BAS:
		  xutil = tabvaleur[i].valeur;
		  yutil = ymin;
		  break;
		case HAUT:
		  xutil = tabvaleur[i].valeur;
		  yutil = ymax;
		  break;
		case GAUCHE:
		  xutil = xmin;
		  yutil = tabvaleur[i].valeur;
		  break;
		case DROITE:
		  xutil = xmax;
		  yutil = tabvaleur[i].valeur;
		  break;
	       }

	       taille = tabvaleur[i].taille;
	       label = tabvaleur[i].label;
	       nsignif = tabvaleur[i].nsignif;
	       Principal = tabvaleur[i].principal;

	       /* if ((i == 0) || (i == ninter)) decale = VRAI; else */
	       decale = FAUX;
	       L3Graduation(sousimage, xutil, yutil,
			    direction, longueur, taille, label,
			    xaxelog, yaxelog, nsignif, decale, facteur,
			    &lastdx, &lastdy, maxdeci, Principal);

	       if ((direction == BAS) || (direction == HAUT))
		  xutil *= facteur;
	       else
		  yutil *= facteur;

/*----------------------------------------------------------------------*
* Determination de xl3, yl3 extreme des labels pour les axes verticaux *
*----------------------------------------------------------------------*/

	       if (i == 0)
		  L3ConversionCoordonnees(xutil, yutil, &xl3, &yl3);
	       else if ((i != ninter) && (label == VRAI))
	       {

		  switch (direction)
		  {
		   case BAS:
		     if (YL3GENTEXTEMIN < yl3)
			yl3 = YL3GENTEXTEMIN;
		     break;
		   case HAUT:
		     if (YL3GENTEXTEMAX > yl3)
			yl3 = YL3GENTEXTEMAX;
		     break;
		   case GAUCHE:
		     if (XL3GENTEXTEMIN < xl3)
			xl3 = XL3GENTEXTEMIN;
		     break;
		   case DROITE:
		     if (XL3GENTEXTEMAX > xl3)
			xl3 = XL3GENTEXTEMAX;
		     break;
		  }

	       }

	    }  /* Fin trace des graduations */

	    free((char *) tabvaleur);
	    L3Axe();

	    /*-------------------*
	     * Ecriture du titre *
	     *-------------------*/

	    switch (direction)
	    {
	     case BAS:
	       xutil = (xmax + xmin) / 2.0;
	       L3ConvCoordL3VersUtilLog(0, yl3, &bidon, &yutil, FAUX, FAUX);
	       yutil -= (ymax - ymin) / 40.0;
	       angle = 0.0;
	       position = POSMH;
	       break;
	     case HAUT:
	       xutil = (xmax + xmin) / 2.0;
	       L3ConvCoordL3VersUtilLog(0, yl3, &bidon, &yutil, FAUX, FAUX);
	       yutil += (ymax - ymin) / 40.0;
	       angle = 0.0;
	       position = POSMB;
	       break;
	     case GAUCHE:
	       L3ConvCoordL3VersUtilLog(xl3, 0, &xutil, &bidon, FAUX, FAUX);
	       xutil -= (xmax - xmin) / 40.0;
	       yutil = (ymax + ymin) / 2.0;
	       angle = 90.0;
	       position = POSMB;
	       break;
	     case DROITE:
	       L3ConvCoordL3VersUtilLog(xl3, 0, &xutil, &bidon, FAUX, FAUX);
	       xutil += (xmax - xmin) / 40.0;
	       yutil = (ymax + ymin) / 2.0;
	       angle = 90.0;
	       position = POSMH;
	       break;
	    }

	    L3ConversionCoordonneesLog(xutil, yutil, &xl3, &yl3, FAUX, FAUX);
	    L3ConvCoordL3VersUtil(xl3, yl3, &xutil, &yutil);

	    strcpy(titrecomplet, titre);

	    if ((*unite != '\0') || (facteur != 1.0))
	    {
	       strcat(titrecomplet, "  (");

	       if (facteur != 1.0)
	       {
		  strcat(titrecomplet, "x");
		  L3BeauLabelPur(facteur, 1, chainefact, -1);
		  strcat(titrecomplet, chainefact);
		  chainenormal[0] = MODNORMAL;
		  chainenormal[1] = '\0';
		  strcat(titrecomplet, chainenormal);

		  if (*unite != '\0')
		     strcat(titrecomplet, " ");

	       }

	       strcat(titrecomplet, unite);
	       strcat(titrecomplet, ")");
	    }

	    L3GenTextePos(sousimage, titrecomplet, xutil, yutil, 0.75,
			    AspectTexte, angle, VRAI, position);

	 }
	 else
	 {
	     L3Axe();
	     fprintf(stderr, "L3AxeGradue : espace memoire insuffisant.\n");
	 }

      }
      else
      {
	  L3Axe();
      }

   }

}

/**********************************
 * Procedure interne : L3Graduation
 *---------------------------------
 * Parametres :
 *   sousimage    : numero de la sous-image concernee
 *   xutil, yutil : coordonnees utilisateur de la graduation
 *   direction    : indicateur de l'axe a tracer
 *   longueur     : longueur maximale de la graduation
 *   taille       : facteur de taille de la graduation
 *   label        : indicateur de presence de la valeur
 *   xaxelog      : indicateur de mode logarithmique sur x
 *   yaxelog      : indicateur de mode logarithmique sur y
 *   nsignif      : nombre de chiffres significatifs pour la valeur
 *   decale       : indicateur de decalage de la valeur
 *   facteur      : facteur multiplicatif pour la factorisation

 * Valeur de retour : aucune

 * Description : trace une graduation

 * Erreurs : aucune
 *****************************************************************/


static void
L3Graduation(
	     int sousimage,
	     REEL xutil,
	     REEL yutil,
	     int direction,
	     REEL longueur,
	     double taille,
	     int label,
	     int xaxelog,
	     int yaxelog,
	     int nsignif,
	     int decale,
	     double facteur,
 /* Encombrement du dernier label */
	     int *lastdx,
	     int *lastdy,
 /* Nombre de decimale des graduations */
	     int Impose,
	     int Principal

)
{
   int             xlog = xaxelog, ylog = FAUX;
   int             xl3, yl3;
   int             position = POSMB;
   REEL            xdebut, ydebut, xfin, yfin;
   REEL            xlabel, ylabel;
   double          valeur = xutil;
   double          decalage;
   char            texte[256];
   /* dit si l'axe est a tracer comme axe supportant un label */
   int             atracer = TRUE;
   double          tailext;
   
   /* Calcul des diverses coordonnees */
   switch (direction) {
    case BAS:
      xlog = xaxelog;
      ylog = FAUX;
      valeur = xutil;
      xdebut = xfin = xutil * facteur;
      ydebut = yutil + taille * longueur;
      yfin = yutil;
      break;
    case HAUT:
      xlog = xaxelog;
      ylog = FAUX;
      valeur = xutil;
      xdebut = xfin = xutil * facteur;
      ydebut = yutil - taille * longueur;
      yfin = yutil;
      break;
    case GAUCHE:
      xlog = FAUX;
      ylog = yaxelog;
      valeur = yutil;
      ydebut = yfin = yutil * facteur;
      xdebut = xutil + taille * longueur;
      xfin = xutil;
      break;
    case DROITE:
      xlog = FAUX;
      ylog = yaxelog;
      valeur = yutil;
      ydebut = yfin = yutil * facteur;
      xdebut = xutil - taille * longueur;
      xfin = xutil;
      break;
   }

   /*--------------------*/
   /* Trace de la valeur */
   /*--------------------*/

   if (label != FAUX) {

      decalage = 1.0;
      xlabel = xutil;
      ylabel = yutil;

      switch (direction) {
       case BAS:
	 xlabel *= facteur;
	 ylabel -= longueur / 200.0 * (2.0 + 1.5 * decalage);
	 position = POSMH;
	 break;
       case HAUT:
	 xlabel *= facteur;
	 ylabel += longueur / 100.0 * (1.5 * decalage);
	 position = POSMB;
	 break;
       case GAUCHE:
	 xlabel -= longueur / 100.0 * (1.5 * decalage);
	 ylabel *= facteur;
	 position = POSDM;
	 break;
       case DROITE:
	 xlabel += longueur / 100.0 * (1.5 * decalage);
	 ylabel *= facteur;
	 position = POSGM;
	 break;
      }

      L3ConversionCoordonneesLog(xlabel, ylabel, &xl3, &yl3, xlog, ylog);

      SETSOUSIMAGE(sousimage);

      if (
	  (xlog && ((direction == BAS) || (direction == HAUT))) ||
	  (ylog && ((direction == DROITE) || (direction == GAUCHE)))
	 ) {
	 /* Cas d'un axe log -> notation scientifique imposee */
	 L3BeauLabelScientifique(valeur, nsignif, texte, Impose);
      } else {
	 /* cas lineaire pas de notation scientifique obligatoire */
	 L3BeauLabelPur(valeur, nsignif, texte, Impose);
      }

      if (TYPEPOLICE == POLICESOFT) {
	 L3ConvCoordL3VersUtil(xl3, yl3, &xlabel, &ylabel);
	 /* Evaluation de l'encombrement */
	 switch (direction) {
	  case BAS:
	  case HAUT:
	    atracer = (xl3 > *lastdx);
	    break;

	  case GAUCHE:
	  case DROITE:
	    atracer = (yl3 > *lastdy);
	    break;
	 }
	 if (atracer) {
	    int             dx, dy;
	    L3TailleTexteNumerique(sousimage, texte, &dx, &dy);
	    L3GenTextePos(sousimage, texte, xlabel, ylabel, 1.0, AspectTexte,
			  0.0, VRAI, position);
	    *lastdx = xl3 + dx;
	    *lastdy = yl3 + dy;
	 }
      } else {
	 int             dx, dy, OldAngle = ANGLETEXTE;

	 ANGLETEXTE = 0.0;
	 L3TailleTexteNumerique(sousimage, texte, &dx, &dy);

	 switch (direction) {
	  case BAS:
	    xl3 -= (dx / 2);
	    yl3 -= dy;
	    break;
	  case HAUT:
	    xl3 -= (dx / 2);
	    break;

	  case GAUCHE:
	    xl3 -= dx;
	    yl3 -= (dy / 2);
	    break;
	  case DROITE:
	    xl3 += (dx / 2);
	    yl3 -= (dy / 2);
	    break;
	 }
	 /* limitation pour les effets de bords */
	 xl3 = (xl3 < XPAGEMIN) ? XPAGEMIN + 2 : xl3;
	 yl3 = (yl3 < YPAGEMIN) ? YPAGEMIN + 2 : yl3;

	 /* Evaluation de l'encombrement */
	 switch (direction) {
	  case BAS:
	  case HAUT:
	    atracer = (xl3 > *lastdx);
	    break;

	  case GAUCHE:
	  case DROITE:
	    atracer = (yl3 > *lastdy);
	    break;
	 }
	 if (atracer) {
	    L3ConvCoordL3VersUtil(xl3, yl3, &xlabel, &ylabel);
	    L3PointCourant(sousimage, xlabel, ylabel);
	    L3TexteNumerique(sousimage, texte);
	    *lastdx = xl3 + dx;
	    *lastdy = yl3 + dy;
	 }
	 ANGLETEXTE = OldAngle;
      }
   }
   /*------------------*/
   /* Trace du segment */
   /*------------------*/
   if (atracer) 
   {
      tailext = ((taille >= TAILLEGRADPRI) && label)? TAILLEGRADPRI: 0.0;
      
      switch (direction) {
       case BAS:
	 yfin = yutil - tailext * longueur / 2.0;
	 break;
       case HAUT:
	 yfin = yutil + tailext * longueur / 2.0;
	 break;
       case GAUCHE:
	 xfin = xutil - tailext * longueur / 2.0;
	 break;
       case DROITE:
	 xfin = xutil + tailext * longueur / 2.0;
	 break;
      }
   }
   
   if ((L3SansTiret) && (taille == TAILLEGRADINT) && (label == FAUX)) {
   } else {
      L3SegmentLog(sousimage, &xdebut, &ydebut, &xfin, &yfin, 1, xlog, ylog);
   }
}

/* Fin de Fichier */
