/* identification du source */

/* "%W% %G% Pierre Brochard" */

/*********************************************************/
/*       sphere 3d parametrable                          */
/*********************************************************/

#ifndef SPXPREC4D
void
SPXPREC(ObjetSphere)(GEN ox,GEN oy,GEN oz,GEN ray,
		GEN valc,unsigned def,SPXFILE* spx_file)
{
	const double pi = M_PI;
	const double pi2 = 2.0*pi;
	const unsigned nx = def;
	const unsigned ny = 2*def;
	const unsigned npoints = nx*ny;
	GEN* RESTRICT const x = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const y = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const z = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const c = (GEN*) malloc(npoints*sizeof(GEN));

	register unsigned j = 0,k = 0;

	if(!x || !y || !z || !c)
	{
	    fprintf(stderr,"SpxObjetSphere> Probleme d'allocation memoire");

	    if(x) free((char*) x);
	    if(y) free((char*) y);
	    if(z) free((char*) z);
	    if(c) free((char*) c);

	    return;
	}

	{
	    register const double pinx = pi/(nx-1);
	    register const double piny = pi2/(ny-1);

#ifdef CRAY
#pragma _CRI ivdep
#endif
	    for(;j<ny;j++)
	    {
		register unsigned i=0;

		for(;i<nx;i++,k++)
		{
		    register const double txx = i*pinx;
		    register const double tyy = j*piny;
		    register const GEN sxx = sin(txx);

		    x[k] = ray*sxx*cos(tyy) + ox;
		    y[k] = ray*sxx*sin(tyy) + oy;
		    z[k] = ray*cos(txx) + oz;
		    c[k] = valc;
		}

	    }

	}

	if(spx_file)
	{
	    SPXPREC(Tetra4d)(nx,ny,x,nx,y,nx,z,nx,c,nx,spx_file);
	}

	free((char*) x);
	free((char*) y);
	free((char*) z);
	free((char*) c);
}
#endif

/*********************************************************/
/*       cylindre 3d parametrable                        */
/*********************************************************/

void
#ifdef SPXPREC4D
/* valc est de dim def. */
SPXPREC(ObjetCylindre4d)(GEN ox1,GEN oy1,GEN oz1,GEN decal1,
		GEN ox2,GEN oy2,GEN oz2,GEN decal2,GEN ray,
		const GEN* RESTRICT const valc,unsigned def,SPXFILE* spx_file)
#else
SPXPREC(ObjetCylindre)(GEN ox1,GEN oy1,GEN oz1,GEN decal1,
		GEN ox2,GEN oy2,GEN oz2,GEN decal2,GEN ray,
		GEN valc,unsigned def,SPXFILE* spx_file)
#endif
{
	const double pi = M_PI;
	const double pi2 = 2.0*pi;
	const unsigned nx = def;
	const unsigned ny = 4;
	const unsigned npoints = nx*ny;
	GEN* RESTRICT const x = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const y = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const z = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const c = (GEN*) malloc(npoints*sizeof(GEN));
	double dx = ox2 - ox1;
	double dy = oy2 - oy1;
	double dz = oz2 - oz1;
	GEN dd = sqrt(dx*dx+dy*dy+dz*dz);
	GEN x1 = ox1,y1t = oy1,z1 = oz1,x2 = ox2,y2 = oy2,z2 = oz2;
	double dxy;
	double dxy2;
	GEN rau;

	register unsigned i = 0;
	register unsigned j = nx;
	register unsigned k = 2*nx;
	register unsigned l = 3*nx;

	if(!x || !y || !z || !c)
	{
	    fprintf(stderr,"SpxObjetCylindre> "
			    "Probleme d'allocation memoire");

	    if(x) free((char*) x);
	    if(y) free((char*) y);
	    if(z) free((char*) z);
	    if(c) free((char*) c);

	    return;
	}

	if(dd == 0.0)
	    dd = 1.;

	dx /= dd;
	dy /= dd;
	dz /= dd;
	x1 += decal1*dx;
	y1t += decal1*dy;
	z1 += decal1*dz;
	x2 -= decal2*dx;
	y2 -= decal2*dy;
	z2 -= decal2*dz;
	dxy2 = dx*dx+dy*dy;
	dxy = sqrt(dxy2);

	if(dxy == 0.0)
	{
	    dx = 1.,dy = 0.,dxy = 1.;
	}

	rau = ray/dxy;

#ifdef CRAY
#pragma _CRI ivdep
#endif
	for(;i<nx;i++,j++,k++,l++)
	{
	    register const double alp = i*pi2/(nx-1);
	    register const GEN calp = cos(alp);
	    register const GEN salp = sin(alp);
	    register const GEN txx = -rau*(calp*dy + salp*dx*dz);
	    register const GEN tyy = rau*(calp*dx - salp*dy*dz);
	    register const GEN tzz = rau*salp*dxy2;

	    x[i] = x1;
	    y[i] = y1t;
	    z[i] = z1;
#ifndef SPXPREC4D
	    c[i] = valc;
#else
	    c[i] = valc[i];
#endif
	    x[j] = x1 + txx;
	    y[j] = y1t + tyy;
	    z[j] = z1 + tzz;
#ifndef SPXPREC4D
	    c[j] = valc;
#else
	    c[j] = valc[i];
#endif
	    x[k] = x2 + txx;
	    y[k] = y2 + tyy;
	    z[k] = z2 + tzz;
#ifndef SPXPREC4D
	    c[k] = valc;
#else
	    c[k] = valc[i];
#endif
	    x[l] = x2;
	    y[l] = y2;
	    z[l] = z2;
#ifndef SPXPREC4D
	    c[l] = valc;
#else
	    c[l] = valc[i];
#endif
	}

	if(spx_file)
	{
	    SPXPREC(Tetra4d)(nx,ny,x,nx,y,nx,z,nx,c,nx,spx_file);
	}

	free((char*) x);
	free((char*) y);
	free((char*) z);
	free((char*) c);
}

/*********************************************************/
/*       collection de cercles 3d                        */
/*********************************************************/

#ifndef SPXPREC4D
void
SPXPREC(CollectionCercles)(const GEN* RESTRICT const ox,
		const GEN* RESTRICT const oy,const GEN* RESTRICT const oz,
		const GEN* RESTRICT const oc,unsigned nval,GEN rmin,GEN rmax,
		unsigned def,SPXFILE* spx_file)
{
    /* Le rayon des cercles est une fonction affine de oz qui le ramene entre
       rmin et rmax.
       si rmin = rmax = 0. alors le rayon n'est pas touche (pas ramene entre rmin
       et rmax). */
 
	const GEN Bmax = MAXFLOAT;
	const double pi = M_PI;
	const double pi2 = 2.0*pi;
	const unsigned npoints = def*nval;
	const unsigned nmailles = nval;
	const unsigned dimmailles = (1+def)*nval;
	GEN* RESTRICT const x = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const y = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const z = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const c = (GEN*) malloc(npoints*sizeof(GEN));
	int* RESTRICT const mailles = (int*) malloc(dimmailles*sizeof(int));
	GEN zmin = Bmax,zmax = -Bmax;
	GEN poids = 1.;
	GEN decal = 0.;
	register unsigned i = 0,j = 0,k = 0;

	if(!x || !y || !z || !c || !mailles)
	{
	    fprintf(stderr,"SpxCollectionCercles> "
			    "Probleme d'allocation memoire");

	    if(x) free((char*) x);
	    if(y) free((char*) y);
	    if(z) free((char*) z);
	    if(c) free((char*) c);
	    if(mailles) free((char*) mailles);

	    return;
	}

    /* points */

	for(j=0;j<nval;j++)
	{
	    register GEN ozj = oz[j];

	    zmin = SpxMin(zmin,ozj);
	    zmax = SpxMax(zmax,ozj);
	}

	if(rmin > rmax)
	{
	    register GEN rmmm = rmin;

	    rmin = rmax;
	    rmax = rmmm;
	}

	if(rmin != 0. || rmax != 0.)
	{

	    if(zmin == zmax)
	    {
		poids = rmax/zmax;
	    }
	    else
	    {
		poids = (rmin-rmax)/(zmin-zmax);
		decal = rmax - zmax*poids;
	    }
					      
	}

	{
	    register const double pidef = pi2/def;

#ifdef CRAY
#pragma _CRI ivdep
#endif
	    for(i=0;i<def;i++)
	    {
		register const double teta = i*pidef;
		register const GEN cteta = cos(teta);
		register const GEN steta = sin(teta);

		for(j=0;j<nval;j++,k++)
		{
		    register const GEN ray = oz[j]*poids + decal;

		    x[k] = ox[j] + ray*cteta;
		    y[k] = oy[j] + ray*steta;
		    z[k] = oz[j];
		    c[k] = oc[j];
		}

	    }

	}

    /* tableau de mailles */

	for(i=0;i<nmailles;i++)
	{
	    mailles[i] = def*(i+1);
	}

	for(i=0;i<def;i++)
	{
	    register const unsigned nvali = i*nval;
	    register int* const pmailles = mailles+nmailles+i;

	    for(j=0;j<nmailles;j++)
	    {
		pmailles[def*j] = nvali+j;
	    }

	}

    /* envoi non structure */

	if(spx_file)
	{
	    SPXPREC(NonStruc4d)(npoints,x,y,z,c,nmailles,dimmailles,
			    mailles,spx_file);
	}

	free((char*) x);
	free((char*) y);
	free((char*) z);
	free((char*) c);
	free((char*) mailles);
}
#endif

/*********************************************************/
/*       collection de vecteurs 3d                       */
/*********************************************************/

#ifndef SPXPREC4D
void
SPXPREC(CollectionVecteurs)(const GEN* RESTRICT const ox,
		const GEN* RESTRICT const oy,const GEN* RESTRICT const oz,
		const GEN* RESTRICT const oc,const GEN* RESTRICT const vx,
		const GEN* RESTRICT const vy,const GEN* RESTRICT const vz,
		unsigned nval,GEN rmin,GEN rmax,SPXFILE* spx_file)
{

    /* La direction du vecteur est vx, vy, vz dont la norme est ramene entre
       rmin et rmax proportionellement a elle
       si rmin = rmax = 0. alors la norme n'est pas touche (pas ramene entre rmin
       et rmax) */

	const GEN Bmax = MAXFLOAT;
	const unsigned npoints = 4*nval;
	const unsigned nmailles = 3*nval;
	const unsigned dimmailles = 3*nmailles;
	GEN* RESTRICT const x = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const y = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const z = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const c = (GEN*) malloc(npoints*sizeof(GEN));
	int* RESTRICT const mailles = (int*) malloc(dimmailles*sizeof(int));
	double lmin = Bmax,lmax = -Bmax;
	double poids = 1.;
	double decal = 0.;
	const double rac2s2 = sqrt(2.0)/2.0;
	const double rac2s20 = 0.1*rac2s2;
	const double r1ac2s20 = 1.-rac2s20;
	const double rmaxsrac3 = rmax/sqrt(3.);
	register unsigned i = 0,j = 0,k = 0,k1 = 0,k2 = 0,k3 = 0;

	if(!x || !y || !z || !c || !mailles)
	{
	    fprintf(stderr,"SpxCollectionVecteurs> "
			    "Probleme d'allocation memoire");

	    if(x) free((char*) x);
	    if(y) free((char*) y);
	    if(z) free((char*) z);
	    if(c) free((char*) c);
	    if(mailles) free((char*) mailles);

	    return;
	}

    /* points */

	for(j=0;j<nval;j++)
	{
	    register const double vxx = vx[j];
	    register const double vyy = vy[j];
	    register const double vzz = vz[j];
	    register const double vrj = sqrt(vxx*vxx+vyy*vyy+vzz*vzz);

	    lmin = SpxMin(lmin,vrj);
	    lmax = SpxMax(lmax,vrj);
	}

	if(rmin > rmax)
	{
	    register const GEN rmmm = rmin;

	    rmin = rmax;
	    rmax = rmmm;
	}

	if(rmin != 0. || rmax != 0.)
	{

	    if(lmin == lmax)
	    {
		poids = rmax/lmax;
	    }
	    else
	    {
		poids = (rmin-rmax)/(lmin-lmax);
		decal = rmax - lmax*poids;
	    }

	}

	for(j=0,k1=nval,k2=2*nval,k3=3*nval;j<nval;j++,k1++,k2++,k3++)
	{
	    register const GEN oxx = ox[j];
	    register const GEN oyy = oy[j];
	    register const GEN ozz = oz[j];
	    register const GEN occ = oc[j];
	    register double vxx = vx[j];
	    register double vyy = vy[j];
	    register double vzz = vz[j];
	    register const double vrj = sqrt(vxx*vxx+vyy*vyy+vzz*vzz);
	    register const double vrjn = vrj*poids + decal;

	    if(vrj > vrjn/MAXDOUBLE)
	    {
		register const double rescal = vrjn/vrj;

		vxx *= rescal;
		vyy *= rescal;
		vzz *= rescal;
	    }
	    else
	    {
		vxx = rmaxsrac3;
		vyy = rmaxsrac3;
		vzz = rmaxsrac3;
	    }

	    x[k1] = oxx;
	    y[k1] = oyy;
	    z[k1] = ozz;
	    c[k1] = occ;
	    x[j] = oxx + vxx;
	    y[j] = oyy + vyy;
	    z[j] = ozz + vzz;
	    c[j] = occ;
	    x[k2] = oxx + vxx*r1ac2s20 + vyy*rac2s20;
	    y[k2] = oyy + vyy*r1ac2s20 - vxx*rac2s20;
	    z[k2] = ozz + vzz*r1ac2s20;
	    c[k2] = occ;
	    x[k3] = oxx + vxx*r1ac2s20 - vyy*rac2s20;
	    y[k3] = oyy + vyy*r1ac2s20 + vxx*rac2s20;
	    z[k3] = ozz + vzz*r1ac2s20;
	    c[k3] = occ;
	}

    /* tableau de mailles */
    /* bug = le compilateur SUNOS_5 SC3.0 foire sur la boucle suivante */
    /* en -O3 ... : un label dans le code assembleur est mauvais */

	for(i=0;i<nmailles;i++)
	{
	    mailles[i] = 2*(i+1);
	}

	for(j=nmailles,k=0;k<nval;k++,j+=6)
	{
	    mailles[j+4] = mailles[j+2] = mailles[j] = k;
	    mailles[j+1] = k + nval;
	    mailles[j+3] = k + 2*nval;
	    mailles[j+5] = k + 3*nval;
	}

    /* envoi non structure */

	if(spx_file)
	{
	    SPXPREC(NonStruc4d)(npoints,x,y,z,c,nmailles,dimmailles,
			    mailles,spx_file);
	}

	free((char*) x);
	free((char*) y);
	free((char*) z);
	free((char*) c);
	free((char*) mailles);
}
#endif

/*********************************************************/
/*       collection de spheres 3d ou 4d                  */
/*********************************************************/

void
#ifndef SPXPREC4D
SPXPREC(CollectionSpheres)(const GEN* RESTRICT const ox,
#else
/*       collection de spheres 4d avec c de dim 2*def*def*nval */
SPXPREC(CollectionSpheres4d)(const GEN* RESTRICT const ox,
#endif
		const GEN* RESTRICT const oy,const GEN* RESTRICT const oz,
		const GEN* RESTRICT const oc,const GEN* RESTRICT const ray,
		unsigned nval,GEN rmin,GEN rmax,unsigned def,SPXFILE* spx_file)
{
    /* Les rayons des spheres sont le tableau ray qui est ramene entre
       rmin et rmax 
       si rmin = rmax = 0. alors ray n'est pas touche (pas ramene entre rmin
       et rmax) */

	const GEN Bmax = MAXFLOAT;
	const double pi = M_PI;
	const double pi2 = 2.0*pi;
	const unsigned nx = def;
	const unsigned ny = 2*def;
	const unsigned npoints = nx*ny*nval;
	const unsigned nmailles = (ny-1)*(nx-1)*nval;
	const unsigned dimmailles = 5*nmailles;
	GEN* RESTRICT const x = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const y = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const z = (GEN*) malloc(npoints*sizeof(GEN));
	GEN* RESTRICT const c = (GEN*) malloc(npoints*sizeof(GEN));
	int* RESTRICT const mailles = (int*) malloc(dimmailles*sizeof(int));
	GEN raymin = Bmax,raymax = -Bmax;
	GEN poids = 1.;
	GEN decal = 0.;
	register unsigned i = 0,j = 0,k = 0,l = 0;

	if(!x || !y || !z || !c || !mailles)
	{
#ifndef SPXPREC4D
	    fprintf(stderr,"SpxCollectionSpheres> "
			    "Probleme d'allocation memoire");
#else
	    fprintf(stderr,"SpxCollectionSpheres4d> "
			    "Probleme d'allocation memoire");
#endif

	    if(x) free((char*) x);
	    if(y) free((char*) y);
	    if(z) free((char*) z);
	    if(c) free((char*) c);
	    if(mailles) free((char*) mailles);

	    return;
	}

    /* points */

	for(j=0;j<nval;j++)
	{
	    register GEN rayj = ray[j];

	    raymin = SpxMin(raymin,rayj);
	    raymax = SpxMax(raymax,rayj);
	}

	if(rmin > rmax)
	{
	    register GEN rmmm = rmin;

	    rmin = rmax;
	    rmax = rmmm;
	}

	if(rmin != 0. || rmax != 0.)
	{

	    if(raymin == raymax)
	    {
		poids = rmax/raymax;
	    }
	    else
	    {
		poids = (rmin-rmax)/(raymin-raymax);
		decal = rmax - raymax*poids;
	    }
					      
	}

	{
	    register const double pinx = pi/(nx-1);
	    register const double piny = pi2/(ny-1);

#ifdef CRAY
#pragma _CRI ivdep
#endif
	    for(k=0,l=0;k<nval;k++)
	    {

#ifdef CRAY
#pragma _CRI ivdep
#endif
		for(j=0;j<ny;j++)
		{

		    for(i=0;i<nx;i++,l++)
		    {
			register const double txx = i*pinx;
			register const double tyy = j*piny;
			register const GEN sxx = sin(txx);
			register const GEN rayred = ray[k]*poids+decal;

			x[l] = rayred*sxx*cos(tyy) + ox[k];
			y[l] = rayred*sxx*sin(tyy) + oy[k];
			z[l] = rayred*cos(txx) + oz[k];
#ifndef SPXPREC4D
			c[l] = oc[k];
#else
                        c[l] = oc[l];
#endif
		    }

		}

	    }

	}

    /* tableau de mailles */

	for(i=0;i<nmailles;i++)
	{
	    mailles[i] = 4*(i+1);
	}

	{
	    register int* const pmailles = mailles+nmailles;
	    register const unsigned nx1 = nx-1;
	    register const unsigned nxy = nx*ny;
	    register const unsigned nxy1 = (nx-1)*(ny-1);

	    for(i=0;i<nmailles;i++)
	    {
		register const unsigned isphere = i/nxy1;
		register const unsigned ij = i%nxy1;
		register const unsigned ii = ij%nx1;
		register const unsigned jj = ij/nx1;
		register const int pm0 = isphere*nxy + ii + jj*nx;
		register const int pm1 = pm0+ 1;
		register const int pm2 = pm1 + nx;
		register const int pm3 = pm2 - 1;
		register const unsigned i4 = 4*i;

		pmailles[i4] = pm0;
		pmailles[i4+1] = pm1;
		pmailles[i4+2] = pm2;
		pmailles[i4+3] = pm3;
	    }

	}

    /* envoi non structure */

	if(spx_file)
	{
	    SPXPREC(NonStruc4d)(npoints,x,y,z,c,nmailles,dimmailles,
			    mailles,spx_file);
	}

	free((char*) x);
	free((char*) y);
	free((char*) z);
	free((char*) c);
	free((char*) mailles);
}
